#!/usr/bin/env python3

"""
Command-line WebSocket client for use with Mod9 ASR WebSocket server.

The first positional argument is the URI of the WebSocket server to connect to.
The second positional argument may specify JSON-formatted request options.

Audio data may be read from stdin and sent to the specified WebSocket server.
Meanwhile, messages received from the WebSocket server are printed on stdout.

Example usage with a pre-recorded file:
```
curl -sL mod9.io/hi.wav | client.py ws://localhost:9980 > results.jsonl
```

Example usage with live streaming audio:
```
sox -dqV1 -twav -r16000 -c1 -b16 - | client.py wss://mod9.io '{"partial":true}'
```


"""

import argparse
import asyncio
import sys

import websockets

# This is a fairly small default, set as a lowest common denominator for all use cases.  The ASR
# Engine receives data with a network buffer size of 128 bytes, but also utilizes an internal audio
# buffer size which is reflected by its "latency" request option. For example, at a default latency
# of 0.24s: 16kHZ 16-bit linear PCM audio would buffer as 0.24 * 16000 * 16/8 = 7680 bytes.
# At 0.10s:  8kHz  8-bit mu-law PCM audio would buffer as 0.10 * 8000 * 8/8 = 800 bytes.
# As general guidance: this parameter can be much larger for batch processing of pre-recorded audio,
# or if a real-time streaming application appears to be incurring significant networking overhead.
AUDIO_MESSAGE_SIZE = 128


async def from_stdin_to_engine(websocket, message_size=AUDIO_MESSAGE_SIZE):
    """
    Read chunked audio from stdin and send as messages to Engine WebSocket.
    Return when stdin is closed, after sending an empty message.

    Args:
        websocket (websockets.server.WebSocketServerProtocol):
            Engine WebSocket to send to (expects empty-message termination).
        message_size (int):
            Read audio chunks of this size; affects streaming latency.

    Returns:
        None
    """
    # Async read from stdin: https://stackoverflow.com/q/64303607/281536
    reader = asyncio.StreamReader()
    protocol = asyncio.StreamReaderProtocol(reader)
    await asyncio.get_event_loop().connect_read_pipe(lambda: protocol, sys.stdin.buffer)

    while True:
        # Read audio bytes from stdin, chunked at the message size.
        audio_chunk = await reader.read(message_size)

        if audio_chunk:
            # Send audio chunk as a WebSocket message of the same size.
            await websocket.send(audio_chunk)
        else:
            break

    # Send an empty message to terminate audio data.
    await websocket.send(b'')


async def from_engine_to_stdout(websocket):
    """
    Receive reply messages from Engine WebSocket, and print to stdout.
    Return when the WebSocket connection is closed by the Engine.

    Args:
        websocket (websockets.server.WebSocketServerProtocol):
            WebSocket to receive from (expecting single-line JSON replies).

    Returns:
        None
    """
    while True:
        try:
            reply_message = await websocket.recv()
        except websockets.ConnectionClosedOK:
            break

        reply_json = reply_message.decode()
        print(reply_json)


async def communicate_with_engine(uri, options_json, message_size):
    async with websockets.connect(uri) as websocket:
        # Send JSON-formatted options as initial message.
        options_message = options_json.encode('utf-8')
        await websocket.send(options_message)

        # Send audio to Engine in an async task.
        # NOTE: if the client sends audio for a command that does not expect it, or if the Engine
        #       closes the connection early (e.g. on error), then it may take 10 seconds for the
        #       WebSocket connection to be closed by the Engine.
        # TODO: add logic to this client to await/parse the first Engine reply before proceeding?
        send_audio = asyncio.get_event_loop().create_task(
            from_stdin_to_engine(websocket, message_size),
        )

        # Receive replies from Engine in an async task.
        recv_replies = asyncio.get_event_loop().create_task(
            from_engine_to_stdout(websocket),
        )

        # Wait for the Engine to send all its replies, and then close the connection.
        await recv_replies

        # Cancel any audio remaining to be sent, since the Engine can't accept it anymore.
        send_audio.cancel()


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'uri',
        metavar='URI',
        help='URI of the WebSocket server.',
        default='ws://localhost:9980',
    )
    parser.add_argument(
        'options_json',
        metavar='OPTIONS_JSON',
        nargs='?',
        help='JSON-formatted request options to pass to ASR Engine.',
        default='{}'
    )
    parser.add_argument(
        '--message-size',
        help='Size of audio sent in each WebSocket message, affecting streaming latency.',
        default=AUDIO_MESSAGE_SIZE,
        type=int,
    )
    args = parser.parse_args()

    try:
        asyncio.get_event_loop().run_until_complete(
            communicate_with_engine(
                args.uri,
                args.options_json,
                args.message_size,
            )
        )
    except KeyboardInterrupt:
        sys.exit(130)  # Don't show Python traceback.
    except Exception:
        print(f"ERROR: Unable to communicate with {args.uri} ... check if server is accessible?")
        sys.exit(1)


if __name__ == '__main__':
    main()



"""
A drop-in replacement for Google Cloud STT Python Client Library.

The speech_mod9 module provides a subset of Google Cloud STT
functionality, with additional Mod9-exclusive features, such as phrase
alternatives, and uses subclassed Google objects directly. Input in
Google-type objects (or Google-style dict) is converted internally to
Mod9-style dict. Output is converted from Mod9-style dict to subclassed
Google-style and -type objects.
"""

import sys

import google.cloud.speech_v1p1beta1 as cloud_speech
from google.protobuf import duration_pb2 as duration
import proto

from mod9.asr import common
from mod9.reformat import config

__version__ = config.WRAPPER_VERSION

# Used to compile protobufs into Python classes.
__protobuf__ = proto.module(
    package='mod9.asr.speech_mod9',
    manifest={
        "LongRunningRecognizeMetadata",
        "LongRunningRecognizeRequest",
        "LongRunningRecognizeResponse",
        "RecognitionAudio",                    # Held by subclass
        "RecognitionConfig",                   # Mod9-only attribute(s)
        "RecognitionMetadata",
        "RecognizeRequest",                    # Holds subclass with Mod9-only attribute(s)
        "RecognizeResponse",                   # Holds subclass with Mod9-only attribute(s)
        "SpeakerDiarizationConfig",
        "SpeechContext",
        "SpeechRecognitionAlternative",
        'SpeechRecognitionPhrase',             # Mod9-only message
        'SpeechRecognitionPhraseAlternative',  # Mod9-only message
        "SpeechRecognitionResult",             # Mod9-only attribute(s)
        "StreamingRecognitionConfig",          # Holds subclass with Mod9-only attribute(s)
        "StreamingRecognitionResult",          # Mod9-only attribute(s)
        "StreamingRecognizeRequest",
        "StreamingRecognizeResponse",          # Holds subclass with Mod9-only attribute(s)
        "WordInfo",
    },
)


# Copy Google's types into this namespace.
# Not all are used/implemented by Mod9 ASR; they may be silently ignored.
LongRunningRecognizeMetadata = cloud_speech.LongRunningRecognizeMetadata
LongRunningRecognizeRequest = cloud_speech.LongRunningRecognizeRequest
LongRunningRecognizeResponse = cloud_speech.LongRunningRecognizeResponse
RecognitionAudio = cloud_speech.RecognitionAudio
RecognitionMetadata = cloud_speech.RecognitionMetadata
SpeakerDiarizationConfig = cloud_speech.SpeakerDiarizationConfig
SpeechContext = cloud_speech.SpeechContext
SpeechRecognitionAlternative = cloud_speech.SpeechRecognitionAlternative
StreamingRecognizeRequest = cloud_speech.StreamingRecognizeRequest
WordInfo = cloud_speech.WordInfo


# Subclass Google's types required to extend
#  functionality to include Mod9-only options.


# Mod9-only message
class SpeechRecognitionPhraseBias(proto.Message):
    """
    Phrase alternative model scores. Mod9-only message.

    Attributes:
        am (float):
            The acoustic model score of the phrase, shifted
            so that the most likely phrase has score 0.0.
        lm (float):
            The language model score of the phrase, shifted
            so that the most likely phrase has score 0.0.
    """

    am = proto.Field(proto.FLOAT, number=1)
    lm = proto.Field(proto.FLOAT, number=2)


# Mod9-only message
class SpeechRecognitionPhrase(proto.Message):
    """
    Phrase alternative hypotheses and model scores. Mod9-only message.

    Attributes:
        bias (SpeechRecognitionPhraseBias):
            The acoustic and language model scores.
        phrase (str):
            Phrase text representing the words the user spoke.
    """

    bias = proto.Field(SpeechRecognitionPhraseBias, number=1)
    phrase = proto.Field(proto.STRING, number=2)


# Mod9-only message
class SpeechRecognitionPhraseAlternative(proto.Message):
    """
    Set of phrase alternatives for a portion of the audio. Mod9-only
    message.

    Attributes:
        alternatives (Sequence[SpeechRecognitionPhrase]):
            Set of phrase hypotheses (up to the maximum specified in
            ``max_phrase_alternatives``). These phrase alternatives are
            ordered in terms of accuracy, with the top (first) phrase
            alternative being the most probable, as ranked by the
            recognizer.
        start_time (duration.Duration):
            Time offset relative to the beginning of the audio, and
            corresponding to the start of the spoken phrase.
        end_time (duration.Duration):
            Time offset relative to the beginning of the audio, and
            corresponding to the end of the spoken phrase.
        phrase (str):
            Most likely phrase text representing the words the user
            spoke.
    """

    alternatives = proto.RepeatedField(SpeechRecognitionPhrase, number=1)
    start_time = proto.Field(duration.Duration, number=2)
    end_time = proto.Field(duration.Duration, number=3)
    phrase = proto.Field(proto.STRING, number=4)


class SpeechRecognitionResult(proto.Message):
    """
    OVERRIDE: A speech recognition result corresponding to a portion of
    the audio. Mod9 subclass with Mod9-only attribute ``phrases``.

    Attributes:
        alternatives (Sequence[SpeechRecognitionAlternative]):
            May contain one or more recognition hypotheses (up to the
            maximum specified in ``max_alternatives``). These
            alternatives are ordered in terms of accuracy, with the top
            (first) alternative being the most probable, as ranked by
            the recognizer.
        language_code (str):
            Output only. The
            `BCP-47 <https://www.rfc-editor.org/rfc/bcp/bcp47.txt>`__
            language tag of the language in this result. This language
            code was detected to have the most likelihood of being
            spoken in the audio.
        phrases (Sequence[SpeechRecognitionPhraseAlternative]):
            Sequence of phrase alternatives in increasing time order.
            Mod9-only attribute.
        channel_tag (int):
            Mod9: not available at present.
    """

    alternatives = proto.RepeatedField(SpeechRecognitionAlternative, number=1)
    language_code = proto.Field(proto.STRING, number=5)
    # Mod9-only attribute
    phrases = proto.RepeatedField(SpeechRecognitionPhraseAlternative, number=901)


class RecognizeResponse(proto.Message):
    """
    OVERRIDE: The only message returned to the client by the
    ``Recognize`` method. It contains the result as zero or more
    sequential ``SpeechRecognitionResult`` messages. Mod9 subclass to
    hold Mod9 subclassed SpeechRecognitionResult.

    Attributes:
        results (Sequence[SpeechRecognitionResult]):
            Sequential list of transcription results corresponding to
            sequential portions of audio.
    """

    results = proto.RepeatedField(SpeechRecognitionResult, number=2)


class StreamingRecognitionResult(proto.Message):
    """
    OVERRIDE: A streaming speech recognition result corresponding to a
    portion of the audio that is currently being processed. Mod9
    subclass with Mod9-only attribute ``phrases``.

    Attributes:
        alternatives (Sequence[SpeechRecognitionAlternative]):
            May contain one or more recognition hypotheses (up to the
            maximum specified in ``max_alternatives``). These
            alternatives are ordered in terms of accuracy, with the top
            (first) alternative being the most probable, as ranked by
            the recognizer.
        is_final (bool):
            If ``false``, this ``StreamingRecognitionResult`` represents
            an interim result that may change. If ``true``, this is the
            final time the speech service will return this particular
            ``StreamingRecognitionResult``, the recognizer will not
            return any further hypotheses for this portion of the
            transcript and corresponding audio.
        stability (float):
            An estimate of the likelihood that the recognizer will not
            change its guess about this interim result. Values range
            from 0.0 (completely unstable) to 1.0 (completely stable).
            This field is only provided for interim results
            (``is_final=false``). The default of 0.0 is a sentinel value
            indicating ``stability`` was not set.
        result_end_time (duration.Duration):
            Time offset of the end of this result relative to the
            beginning of the audio.
        language_code (str):
            Output only. The
            `BCP-47 <https://www.rfc-editor.org/rfc/bcp/bcp47.txt>`__
            language tag of the language in this result. This language
            code was detected to have the most likelihood of being
            spoken in the audio.
        phrases (Sequence[SpeechRecognitionPhraseAlternative]):
            Sequence of phrase alternatives in increasing time order.
            Mod9-only attribute.
        channel_tag (int):
            Mod9: not available at present.
    """

    alternatives = proto.RepeatedField(SpeechRecognitionAlternative, number=1)
    is_final = proto.Field(proto.BOOL, number=2)
    stability = proto.Field(proto.FLOAT, number=3)
    result_end_time = proto.Field(duration.Duration, number=4)
    language_code = proto.Field(proto.STRING, number=6)
    # Mod9-only attribute
    phrases = proto.RepeatedField(SpeechRecognitionPhraseAlternative, number=901)


class StreamingRecognizeResponse(proto.Message):
    """
    OVERRIDE: ``StreamingRecognizeResponse`` is the only message
    returned to the client by ``StreamingRecognize``. A series of zero
    or more ``StreamingRecognizeResponse`` messages are streamed back
    to the client. If there is no recognizable audio, and
    ``single_utterance`` is set to false, then no messages are streamed
    back to the client. Mod9 subclass to hold Mod9 subclassed
    ``StreamingRecognitionResult``.

    Here's an example of a series of ten
    ``StreamingRecognizeResponse`` s that might be returned while
    processing audio:

    1. results { alternatives { transcript: "tube" } stability: 0.01 }

    2. results { alternatives { transcript: "to be a" } stability: 0.01
       }

    3. results { alternatives { transcript: "to be" } stability: 0.9 }
       results { alternatives { transcript: " or not to be" } stability:
       0.01 }

    4. results { alternatives { transcript: "to be or not to be"
       confidence: 0.92 } alternatives { transcript: "to bee or not to
       bee" } is_final: true }

    5. results { alternatives { transcript: " that's" } stability: 0.01
       }

    6. results { alternatives { transcript: " that is" } stability: 0.9
       } results { alternatives { transcript: " the question" }
       stability: 0.01 }

    7. results { alternatives { transcript: " that is the question"
       confidence: 0.98 } alternatives { transcript: " that was the
       question" } is_final: true }

    Notes:

    -  Only two of the above responses #4 and #7 contain final results;
       they are indicated by ``is_final: true``. Concatenating these
       together generates the full transcript: "to be or not to be that
       is the question".

    -  The others contain interim ``results``. #3 and #6 contain two
       interim ``results``: the first portion has a high stability and
       is less likely to change; the second portion has a low stability
       and is very likely to change. A UI designer might choose to show
       only high stability ``results``.

    -  The specific ``stability`` and ``confidence`` values shown above
       are only for illustrative purposes. Actual values may vary.
       (Mod9 does not currently support ``stability`` and returns a
       placeholder ``0.5`` value).

    -  In each response, only one of these fields will be set:
       ``error``, ``speech_event_type``, or one or more (repeated)
       ``results``.

    Attributes:
        results (Sequence[StreamingRecognitionResult]):
            This repeated list contains zero or more results that
            correspond to consecutive portions of the audio currently
            being processed. It contains zero or one ``is_final=true``
            result (the newly settled portion), followed by zero or more
            ``is_final=false`` results (the interim results).
        error (status.Status):
            Mod9: not available at present.
        speech_event_type (StreamingRecognizeResponse.SpeechEventType):
            Mod9: not available at present.
    """

    results = proto.RepeatedField(StreamingRecognitionResult, number=2)


class RecognitionConfig(proto.Message):
    """
    OVERRIDE: Provides information to the recognizer that specifies how
    to process the request. Mod9 subclass to support Mod9-only attribute
    ``max_phrase_alternatives``.

    Attributes:
        encoding (RecognitionConfig.AudioEncoding):
            Encoding of audio data sent in all ``RecognitionAudio``
            messages. This field is optional for ``WAV``
            audio files and required for all other audio formats. For
            details, see
            [AudioEncoding][RecognitionConfig.AudioEncoding].
        sample_rate_hertz (int):
            Sample rate in Hertz of the audio data sent in all
            ``RecognitionAudio`` messages. Valid values are 8000, 16000.
            16000 is optimal. For best results, set the sampling rate of
            the audio source to 16000 Hz. If that's not possible, use
            the native sample rate of the audio source (instead of
            re-sampling). This field is optional for WAV audio
            files, but is required for all other audio formats. For
            details, see
            [AudioEncoding][RecognitionConfig.AudioEncoding].
        language_code (str):
            The language of the supplied audio as a
            `BCP-47 <https://www.rfc-editor.org/rfc/bcp/bcp47.txt>`__
            language tag. Example: "en-US". Mod9 currently supports
            ``en-US``.
        max_alternatives (int):
            Maximum number of recognition hypotheses to be returned.
            Specifically, the maximum number of
            ``SpeechRecognitionAlternative`` messages within each
            ``SpeechRecognitionResult``. The server may return fewer
            than ``max_alternatives``. Valid values are ``0``-``30``. A
            value of ``0`` or ``1`` will return a maximum of one. If
            omitted, will return a maximum of one.
        enable_word_time_offsets (bool):
            If ``true``, the top result includes a list of words and the
            start and end time offsets (timestamps) for those words. If
            ``false``, no word-level time offset information is
            returned. The default is ``false``.
        enable_word_confidence (bool):
            If ``true``, the top result includes a list of words and the
            confidence for those words. If ``false``, no word-level
            confidence information is returned. The default is
            ``false``.
        enable_automatic_punctuation (bool):
            If 'true', adds punctuation to recognition result
            hypotheses. This feature is only available in select
            languages. Setting this for requests in other languages
            has no effect at all. The default 'false' value does not add
            punctuation to result hypotheses.
        max_phrase_alternatives (int):
            Mod9-only attribute. Maximum number of phrase hypotheses to
            be returned. Specifically, the maximum number of
            ``SpeechRecognitionPhrase`` messages within each
            ``SpeechRecognitionPhraseAlternative``. The server may
            return fewer than ``max_phrase_alternatives``. Valid values
            are ``0``-``100``. A value of ``0`` or ``1`` will return a
            maximum of one. If omitted, will return a maximum of one.
        latency (float):
            Mod9-only attribute. Chunk size for ASR processing
            (in seconds). Low values increase CPU usage; high
            values degrade endpointing. Default is ``0.24``.
            Valid values are between ``0.01`` and ``3.0``.
        speed (int):
            Mod9-only attribute. Increasing speed will trade-off accuracy
            and diversity of recognition alternatives.  Default is ``5``.
            Valid values are between ``1`` and ``9``.
        options_json (str):
            Mod9-only attribute. Additional request options specified as
            a JSON object. This will override options set by other means.
        audio_channel_count (int):
            Mod9: not available at present.
        enable_separate_recognition_per_channel (bool):
            Mod9: not available at present.
        alternative_language_codes (Sequence[str]):
            Mod9: not available at present.
        profanity_filter (bool):
            Mod9: not available at present.
        adaptation (~.resource.SpeechAdaptation):
            Mod9: not available at present.
        speech_contexts (Sequence[~.cloud_speech.SpeechContext]):
            Mod9: not available at present.
        enable_speaker_diarization (bool):
            Mod9: not available at present.
        diarization_speaker_count (int):
            Mod9: not available at present.
        diarization_config (~.cloud_speech.SpeakerDiarizationConfig):
            Mod9: not available at present.
        metadata (~.cloud_speech.RecognitionMetadata):
            Mod9: not available at present.
        model (str):
            Mod9: not available at present.
        use_enhanced (bool):
            Mod9: not available at present.
    """

    class AudioEncoding(proto.Enum):
        """
        OVERRIDE: The encoding of the audio data sent in the request.
        Mod9 subclass to support Mod9-only attribute ``ALAW``.

        All encodings support only 1 channel (mono) audio.

        For best results, the audio source should be captured and
        transmitted using a lossless encoding (``LINEAR16``). The
        accuracy of the speech recognition can be reduced if lossy
        codecs are used to capture or transmit audio, particularly if
        background noise is present. Lossy codecs include ``ALAW``, and
        ``MULAW``.

        The ``WAV`` audio file formats include a header that describes
        the included audio content. You can request recognition for
        ``WAV`` files that contain either ``LINEAR16``, ``ALAW``, or
        ``MULAW`` encoded audio. If you send ``WAV`` audio file format
        in your request, you do not need to specify an
        ``AudioEncoding``; the audio encoding format is determined from
        the file header. If you specify an ``AudioEncoding`` when you
        send ``WAV`` audio, the encoding configuration must match the
        encoding described in the audio header.
        """

        ENCODING_UNSPECIFIED = 0
        LINEAR16 = 1
        MULAW = 3
        # Mod9-only option
        ALAW = 901

    encoding = proto.Field(proto.ENUM, number=1, enum=AudioEncoding,)
    sample_rate_hertz = proto.Field(proto.INT32, number=2)
    language_code = proto.Field(proto.STRING, number=3)
    max_alternatives = proto.Field(proto.INT32, number=4)
    enable_word_time_offsets = proto.Field(proto.BOOL, number=8)
    model = proto.Field(proto.STRING, number=13)
    enable_word_confidence = proto.Field(proto.BOOL, number=15)
    enable_automatic_punctuation = proto.Field(proto.BOOL, number=11)
    # Mod9-only option
    max_phrase_alternatives = proto.Field(proto.INT32, number=901)
    latency = proto.Field(proto.FLOAT, number=902)
    speed = proto.Field(proto.INT32, number=903)
    options_json = proto.Field(proto.STRING, number=904)


class StreamingRecognitionConfig(proto.Message):
    """
    OVERRIDE: Provides information to the recognizer that specifies how
    to process the request. Mod9 subclass to hold Mod9 subclassed
    ``RecognitionConfig``.

    Attributes:
        config (RecognitionConfig):
            Required. Provides information to the recognizer that
            specifies how to process the request.
        interim_results (bool):
            If ``true``, interim results (tentative hypotheses) may be
            returned as they become available (these interim results are
            indicated with the ``is_final=false`` flag). If ``false`` or
            omitted, only ``is_final=true`` result(s) are returned.
        single_utterance (bool):
            Mod9: not available at present.
    """

    config = proto.Field(RecognitionConfig, number=1)
    interim_results = proto.Field(proto.BOOL, number=3)


class RecognitionAudio(proto.Message):
    """
    OVERRIDE: Contains audio data in the encoding specified in the
    ``RecognitionConfig``. Either ``content`` or ``uri`` must be
    supplied. Mod9 does not limit content size. Mod9 subclass.

    Attributes:
        content (bytes):
            The audio data bytes encoded as specified in
            ``RecognitionConfig``. Note: as with all bytes fields, proto
            buffers use a pure binary representation, whereas JSON
            representations use base64.
        uri (str):
            URI that points to a file that contains audio data bytes as
            specified in ``RecognitionConfig``. The file must not be
            compressed (for example, gzip). Mod9 supports a variety of
            storage backends: ``file://``, ``gs://``, ``http://``,
            ``s3://``.
    """

    content = proto.Field(proto.BYTES, number=1, oneof="audio_source")
    uri = proto.Field(proto.STRING, number=2, oneof="audio_source")


class RecognizeRequest(proto.Message):
    """
    OVERRIDE: The top-level message sent by the client for the
    ``Recognize`` method. Mod9 subclass to hold subclassed objects.

    Attributes:
        config (RecognitionConfig):
            Required. Provides information to the recognizer that
            specifies how to process the request.
        audio (RecognitionAudio):
            Required. The audio data to be recognized.
    """

    config = proto.Field(RecognitionConfig, number=1)
    audio = proto.Field(RecognitionAudio, number=2)


class SpeechClient(object):
    """A duck-typed extension of google.cloud.speech.SpeechClient."""

    def __init__(self, host=None, port=None, *args, **kwargs):
        """Ignore arguments other than host and port, and set Mod9's custom transport."""
        if host is not None:
            config.ASR_ENGINE_HOST = host
        if port is not None:
            config.ASR_ENGINE_PORT = port

    def long_running_recognize(self, request=None, *, config=None, audio=None, **kwargs):
        """
        Performs asynchronous speech recognition.

        Mod9: Not currently implemented.
        """
        # TODO: similar logic to recognize for handling various input types.
        return common.long_running_recognize(request, **kwargs)

    def recognize(self, request=None, *, config=None, audio=None, **kwargs):
        """
        Performs synchronous speech recognition: receive results after
        all audio has been sent and processed.

        Args:
            request (Union[dict, RecognizeRequest]):
                The request object. The top-level message sent by the
                client for the `Recognize` method.
            config (RecognitionConfig):
                Required. Provides information to the recognizer that
                specifies how to process the request. This corresponds
                to the ``config`` field on the ``request`` instance; if
                ``request`` is provided, this should not be set.
            audio (RecognitionAudio):
                Required. The audio data to be recognized. This
                corresponds to the ``audio`` field on the ``request``
                instance; if ``request`` is provided, this should not be
                set.

        Returns:
            RecognizeResponse:
                The only message returned to the client by the
                ``Recognize`` method. It contains the result as zero or
                more sequential ``SpeechRecognitionResult`` messages.
        """

        # Replicate the logic for handling various input types.
        has_flattened_params = any([config, audio])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the ``request`` argument is set, then none of "
                "the individual field arguments should be set."
            )
        if not isinstance(request, RecognizeRequest):
            request = RecognizeRequest(**request)
        return common.recognize(request, module=sys.modules[__name__], **kwargs)

    def streaming_recognize(self, config, requests, **kwargs):
        """
        Perform streaming speech recognition.

        This method allows you to receive results while sending audio.

        Example:
          >>> from mod9.asr import speech_mod9 as speech
          >>> client = speech.SpeechClient()
          >>> encoding = speech.RecognitionConfig.AudioEncoding.LINEAR16
          >>> config = speech.StreamingRecognitionConfig(
          ...     config=speech.RecognitionConfig(
          ...         encoding=encoding,
          ...     ),
          ... )
          >>> # Byte-encoded content goes in ``audio_content`` below:
          >>> request = speech.StreamingRecognizeRequest(audio_content=b'')
          >>> requests = [request]
          >>> for element in client.streaming_recognize(config, requests):
          ...     # process element
          ...     pass

        Args:
            config (StreamingRecognitionConfig):
                The configuration to use for the stream.
            requests (Iterable[StreamingRecognizeRequest]):
                The input objects.

        Returns:
            Iterable[StreamingRecognizeResponse]
        """

        # Replicate the logic of google.cloud.speech_v1.helpers.SpeechHelpers.
        return common.streaming_recognize(
            self._streaming_request_iterable(config, requests),
            module=sys.modules[__name__],
            **kwargs,
        )

    def _streaming_request_iterable(self, config, requests):
        # Replicate the logic of google.cloud.speech_v1.helpers.SpeechHelpers.
        yield {'streaming_config': config}
        for request in requests:
            yield request


# Used to compile protobufs into Python classes.
__all__ = tuple(sorted(__protobuf__.manifest))



#!/usr/bin/env python3
# TODO: describe this file.

import argparse
import json

# from google.protobuf import json_format

from mod9.asr import speech_mod9 as speech
# from mod9.asr import speech


def prepare_dict_test_inputs(content_file_name):
    encoding = speech.RecognitionConfig.AudioEncoding.LINEAR16
    sample_rate_hertz = 8000
    language_code = 'en-US'
    config = {
        'encoding': encoding,
        'sample_rate_hertz': sample_rate_hertz,
        'language_code': language_code,
        # 'enable_word_confidence': True,
        # 'enable_word_time_offsets': True,
        # 'max_alternatives': 2,
        'max_phrase_alternatives': 2,
    }
    with open(content_file_name, 'rb') as content_file:
        content = content_file.read()
    audio = {'content': content}
    return config, audio


# TODO: Remove or upgrade this old code.
# def dict_inputs_to_google_inputs(config, audio):
#     google_config = json_format.ParseDict(config, speech.RecognitionConfig())
#     google_audio = speech.RecognitionAudio(content=audio['content'])
#     return google_config, google_audio


def recognize_test(file_name, speech_client_input_type='dict'):
    client = speech.SpeechClient()
    config_dict, audio_dict = prepare_dict_test_inputs(content_file_name=file_name)
    if speech_client_input_type == 'dict':
        return client.recognize({'config': config_dict, 'audio': audio_dict})
    # elif speech_client_input_type == 'google_types':
    #     config_google, audio_google = dict_inputs_to_google_inputs(config_dict, audio_dict)
    #     return client.recognize({'config': config_google, 'audio': audio_google})
    else:
        raise TypeError(
            "SpeechClient() accepts 'dict' or 'google_types' input. "
            f"'{speech_client_input_type}' not allowed."
        )


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-f', '--file',
        dest='file_name',
        required=True,
        type=str,
        help='Local WAV format audio file to be transcribed.',
    )
    parser.add_argument(
        '-t', '--type',
        dest='input_type',
        default='dict',
        choices=['dict', 'google_types'],
        type=str,
        help="Choose from SpeechClient's allowed input types.",
    )
    args = parser.parse_args()

    response = recognize_test(args.file_name, args.input_type)

    if isinstance(response, dict):
        print(json.dumps(response, indent=2), end='')
    else:
        print(response, end='')



"""
Python SDK and REST API, high-level interfaces to the Mod9 ASR Engine.

Requires a Mod9 ASR Engine server; see https://mod9.io for details.

More documentation and example usage can be found at
  https://mod9.io/python
  https://mod9.io/rest

Within you will find the following modules:

``mod9.asr`` implements the Mod9 ASR Python SDK.
Two versions are offered within this module, both of which are drop-in
replacements for the Google STT Python Client Library"
  - ``mod9.asr.speech``, uses Google's objects for input/output and
    offers a strict subset of Google functionality.
  - ``mod9.asr.speech_mod9`` extends Google's functionality with
     Mod9-exclusive options, such as phrase alternatives.

``mod9.rest.server`` implements the Mod9 ASR REST API, a compatible
drop-in replacement for the Google Cloud STT REST API.
Run ``mod9-asr-rest-api`` from the command line to launch a REST server.
This script can be installed by ``pip --user`` at ``. ~/.local/bin``.

``mod9.reformat`` contains the internal library support. For example,
``mod9.reformat.google`` converts between Mod9 and Google data formats,
while ``mod9.reformat.utils`` communicates with the Mod9 ASR Engine.
"""



#!/usr/bin/env python3

import setuptools

from mod9.reformat import config

with open('README.md') as f_in:
    long_description = f_in.read()

install_requires = [
    'boto3>=1.9.0',
    'elasticsearch>=7.8.1',
    'flask-restful>=0.3.9',
    'google-auth>=1.22.0',
    'google-cloud-speech>=2.0.0',
    'google-resumable-media>=1.0.0',
    'packaging>=15.0',
    'proto-plus>=1.4.0',
    'websockets>=8.0',
]

setuptools.setup(
    name='mod9-asr',
    version=config.WRAPPER_VERSION,
    description='Mod9 ASR Python SDK, REST API, and Websocket Interface:'
                ' high-level interfaces to the Mod9 ASR Engine.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Mod9 Technologies',
    author_email='support@mod9.com',
    license='BSD 2-Clause',
    url='https://github.com/mod9-asr/python-sdk',
    # TODO: classifiers?
    # TODO: platforms?
    packages=setuptools.PEP420PackageFinder.find(exclude=('test',)),
    # TODO: namespace_packages?
    install_requires=install_requires,
    # TODO: extras_require?
    # TODO: python_requires?
    # TODO: scripts?
    # TODO: include_package_data?
    # TODO: zip_safe?

    # Installs executable under user's PATH.
    entry_points={
        'console_scripts': [
            'mod9-asr-elasticsearch-client = mod9.es.client:main',
            'mod9-asr-rest-api = mod9.rest.server:main',
            'mod9-asr-websocket-client = mod9.websocket.client:main',
            'mod9-asr-websocket-server = mod9.websocket.server:main',
        ],
    },
)



"""
Functions for converting Google-style and -type input and output to/from
Mod9-style and -type.
"""

import base64
from binascii import Error as BinasciiError
from collections import OrderedDict
import itertools
import json
import logging
from types import GeneratorType

from mod9.reformat import utils
from mod9.reformat.config import (
    WRAPPER_ENGINE_COMPATIBILITY_RANGE,
    WRAPPER_VERSION,
)

CHUNKSIZE = 8 * 1024


class ConflictingGoogleAudioSettingsError(Exception):
    pass


class Mod9EngineFailedStatusError(Exception):
    pass


# Used to allow arbitrary `model`s to be loaded into Engine and requested by user.
class ObjectContainingEverything:
    """Object that always returns True to `if x in ObjectContainingEverything()` queries."""
    def __contains__(self, x):
        return True


# Used to allow number ranges to be requested by user.
class RealNumericalRangeObject:
    """Object that returns True if ``x`` is in given numerical range."""
    def __init__(self, low, high):
        self.low = low
        self.high = high

    def __contains__(self, x):
        return x >= self.low and x <= self.high


class GoogleConfigurationSettingsAndMappings:
    """
    Hold dicts used to translate keys and values from Google to Mod9
    forms.

    Internally use camelCase for google_allowed_keys. We chose to use
    camelCase internally for compatibility with Google's
    ``protobuf.json_format`` funtions (which are called by ``to_json()``
    and ``from_json()``). We need to make sure we use the proper casing,
    both internally and for output, or can run into non-obvious bugs.
    These bugs can be non-obvious because many of the primary attributes
    are one-word attributes (like ``.transcript``) and so functionality
    will only break if it is dependent on multi-word attributes.

    Input from REST API come in camelCase. Input from Python SDK come in
    snake_case, are converted to camelCase using either ``from_json()``
    (in the case of Google protobuf object input), or
    ``mod9.reformat.utils.recursively_convert_dict_keys_case()`` and
    ``reformat.utils.snake_case_to_camel_case()`` (in the case of dict
    input). Output from REST API use camelCase. Output from Python SDK
    converted to snake_case in ``to_json()``.
    """
    def __init__(self):
        """
        Define dicts used to translate keys and values from Google to
        Mod9 forms.
        """

        # Mod9 and Google input key names for use in following dict's.
        self.mod9_allowed_keys = (
            'encoding',
            'rate',
            'word-confidence',
            'word-intervals',
            'transcript-formatted',
            'transcript-alternatives',
            'language-code',             # transformed to a real Mod9 option
            'model',                     # transformed to a real Mod9 option
            'phrase-alternatives',       # only works in speech_mod9 or REST
            'phrase-alternatives-bias',  # only works in speech_mod9 or REST
            'latency',                   # only works in speech_mod9
            'speed',                     # only works in speech_mod9 or REST
            'options_json',              # only works in speech_mod9 or REST
            'asr-model',                 # Mod9 internal (always happens under the hood)
            'nlp-model',                 # Mod9 internal (always happens under the hood)
        )
        self.google_allowed_keys = (
            'encoding',
            'sampleRateHertz',
            'enableWordConfidence',
            'enableWordTimeOffsets',
            'enableAutomaticPunctuation',
            'maxAlternatives',
            'languageCode',
            'model',
            'maxPhraseAlternatives',     # only works in speech_mod9 or REST
            'enablePhraseConfidence',    # only works in speech_mod9 or REST
            'latency',                   # only works in speech_mod9
            'speed',                     # only works in speech_mod9 or REST
            'optionsJson',               # only works in speech_mod9 or REST
            'asrModel',                  # Mod9 internal (always happens under the hood)
            'nlpModel',                  # Mod9 internal (always happens under the hood)
        )

        # Map from (Google key) to (Mod9 key).
        self.google_to_mod9_key_translations = dict(
            zip(
                self.google_allowed_keys,
                self.mod9_allowed_keys,
            )
        )

        # Allowed Google values.
        self.google_encoding_allowed_values = {
            'ENCODING_UNSPECIFIED',
            'LINEAR16',
            'MULAW',
            'ALAW',  # only in Mod9
        }
        self.google_rate_allowed_values = range(8000, 48001)
        self.google_confidence_allowed_values = {True, False}
        self.google_timestamp_allowed_values = {True, False}
        self.google_punctuation_allowed_values = {True, False}
        self.google_max_alternatives_allowed_values = range(10001)
        self.google_language_code_allowed_values = ObjectContainingEverything()
        self.google_model_allowed_values = {'command_and_search', 'phone_call', 'video', 'default'}
        self.google_max_phrase_alternatives_allowed_values = range(10001)
        self.google_phrase_confidence_allowed_values = {True, False}
        self.latency_allowed_values = RealNumericalRangeObject(0.01, 3.0)
        self.speed_allowed_values = range(1, 10)
        self.options_json_allowed_values = ObjectContainingEverything()
        self.asr_model_allowed_values = ObjectContainingEverything()
        self.nlp_model_allowed_values = ObjectContainingEverything()

        # Group allowed Google values. To be used in following dict.
        self.google_allowed_values = (
            self.google_encoding_allowed_values,
            self.google_rate_allowed_values,
            self.google_confidence_allowed_values,
            self.google_timestamp_allowed_values,
            self.google_punctuation_allowed_values,
            self.google_max_alternatives_allowed_values,
            self.google_language_code_allowed_values,
            self.google_model_allowed_values,
            self.google_max_phrase_alternatives_allowed_values,
            self.google_phrase_confidence_allowed_values,
            self.latency_allowed_values,
            self.speed_allowed_values,
            self.options_json_allowed_values,
            self.asr_model_allowed_values,
            self.nlp_model_allowed_values,
        )

        # Map from (Mod9 key) to (allowed Google values) for given key.
        self.mod9_keys_to_allowed_values = dict(
            zip(
                self.mod9_allowed_keys,
                self.google_allowed_values,
            )
        )

        # Map from (allowed Google encodings) to (Mod9 encodings)
        self.google_encoding_to_mod9_encoding = {
            'LINEAR16': 'pcm_s16le',
            'MULAW': 'mu-law',
            'ALAW': 'a-law',
        }


def input_to_mod9(google_input_settings, module):
    """
    Wrapper method to take Google inputs of various types and return
    Mod9-compatible inputs.

    Args:
        google_input_settings (dict):
            Contains dicts or Google-like-types ``.config`` and
            ``.audio``: options for transcription and audio to be
            transcribed, respectively.
        module (module):
            Module to read Google-like-types from, in case of
            subclassing.

    Returns:
        tuple:
            mod9_config_settings (dict):
                Mod9-style options to pass to Mod9 ASR Engine TCP Server.
            mod9_audio_settings (dict):
                Mod9-style audio to pass to Mod9 ASR Engine TCP Server.
    """

    engine_version = utils.get_version_mod9()
    if not utils.is_compatible_mod9(engine_version):
        raise utils.Mod9IncompatibleEngineVersionError(
            f"Python SDK version {WRAPPER_VERSION} compatible range"
            f" {WRAPPER_ENGINE_COMPATIBILITY_RANGE}"
            f" does not include given Engine of version {engine_version}."
            ' Please use a compatible SDK-Engine pairing. Exiting.'
        )

    # Convert keys from snake_case to camelCase (if necessary), which we use internally.
    #  See docstring for GoogleConfigurationSettingsAndMappings for more info.
    google_input_settings = utils.recursively_convert_dict_keys_case(
        google_input_settings,
        utils.snake_case_to_camel_case,
    )

    # Convert Google-type inputs to dict-type inputs, if necessary.
    if isinstance(google_input_settings['config'], dict):
        google_config_settings_dict = google_input_settings['config']
    elif isinstance(google_input_settings['config'], module.RecognitionConfig):
        google_config_settings_dict = json.loads(
            module.RecognitionConfig.to_json(
                google_input_settings['config'],
                use_integers_for_enums=False,
            )
        )
    elif isinstance(google_input_settings['config'], module.StreamingRecognitionConfig):
        google_config_settings_dict = json.loads(
            module.StreamingRecognitionConfig.to_json(
                google_input_settings['config'],
                use_integers_for_enums=False,
            )
        )

    if 'audio' not in google_input_settings or google_input_settings['audio'] is None:
        # Empty dict will lead to mod9_audio_settings returning None.
        google_audio_settings_dict = dict()
    elif isinstance(google_input_settings['audio'], dict):
        google_audio_settings_dict = google_input_settings['audio']
    elif isinstance(google_input_settings['audio'], module.RecognitionAudio):
        google_audio_settings_dict = json.loads(
            module.RecognitionAudio.to_json(google_input_settings['audio'])
        )

    # Convert Google-style inputs to Mod9-style inputs.
    mod9_config_settings = google_config_settings_to_mod9(google_config_settings_dict)
    mod9_audio_settings = google_audio_settings_to_mod9(google_audio_settings_dict)

    # None is a placeholder since we need to inspect the audio_settings to determine file format.
    if 'format' in mod9_config_settings and mod9_config_settings['format'] is None:
        # Set file type based on file header.
        if isinstance(mod9_audio_settings, GeneratorType):
            # Split generator so utils.parse_wav_encoding() can look at content header.
            mod9_audio_settings, mod9_audio_settings_clone = itertools.tee(mod9_audio_settings)
        wav_encoding, wav_sample_rate = utils.parse_wav_encoding(mod9_audio_settings)
        if wav_encoding:
            mod9_config_settings['format'] = 'wav'
            if 'encoding' in mod9_config_settings:
                if wav_encoding != mod9_config_settings['encoding']:
                    # The Google Cloud STT API complains if WAV and encoding are mismatched.
                    raise ConflictingGoogleAudioSettingsError(
                        "WAV file format encoded as %s should match config specified as %s."
                        % (wav_encoding, mod9_config_settings['encoding'])
                    )
                # The Mod9 ASR Engine complains if both WAV format and audio encoding are specified.
                del mod9_config_settings['encoding']
            if wav_sample_rate:
                if 'rate' in mod9_config_settings:
                    if mod9_config_settings['rate'] != wav_sample_rate:
                        raise ValueError(f"Specified rate, {mod9_config_settings['rate']},"
                                         f" differs from WAV header rate, {wav_sample_rate}.")
                else:
                    mod9_config_settings['rate'] = wav_sample_rate
        else:
            mod9_config_settings['format'] = 'raw'
            if 'encoding' not in mod9_config_settings:
                raise KeyError('Must specify an audio encoding for non-WAV file formats')
        # Reset audio using tee'd clone if it exists.
        try:
            mod9_audio_settings = mod9_audio_settings_clone
        except UnboundLocalError:
            pass

    # Mod9 TCP does not accept 'rate' argument for 'wav' format.
    if 'format' not in mod9_config_settings or mod9_config_settings['format'] == 'wav':
        if 'rate' in mod9_config_settings:
            del mod9_config_settings['rate']

    # Need transcript intervals to mirror Google response format.
    mod9_config_settings['transcript-intervals'] = True

    return mod9_config_settings, mod9_audio_settings


def google_config_settings_to_mod9(google_config_settings):
    """
    Map from Google-style key:value inputs to Mod9 ASR TCP server-style
    key:value inputs.

    Args:
        google_config_settings (dict):
            Google-style options.

    Returns:
        dict:
            Mod9-style options to pass to Mod9 ASR Engine TCP Server.
    """

    settings = GoogleConfigurationSettingsAndMappings()
    mod9_config_settings = dict()

    # StreamingRecognitionConfig has config attribute of type RecognitionConfig.
    if 'config' in google_config_settings:
        # Grab streaming options and assign google_config_settings to RecognitionConfig attribute.
        if google_config_settings.get('singleUtterance') is True:
            raise NotImplementedError(
                'Streaming recognize not yet implemented for Google option single_utterance: True.'
            )
        mod9_config_settings['partial'] = google_config_settings.get('interimResults', False)
        google_config_settings = google_config_settings['config']

        # Turn off batch mode for streaming, otherwise partial will not work.
        mod9_config_settings['batch-threads'] = 0
    else:
        # Use max number of threads available for best speed.
        mod9_config_settings['batch-threads'] = -1

        # This option only applies with speech_mod9.
        if 'latency' in google_config_settings:
            raise KeyError("Option key 'latency' only supported for streaming requests.")

    if 'languageCode' not in google_config_settings:
        raise KeyError("Config missing required key 'languageCode'/'language_code'.")

    # ``to_json()`` populates absent attributes of config with falsy values -> exceptions later.
    google_config_settings = {
        key: value for key, value in google_config_settings.items() if value
    }

    # A subset of possible Google keys are supported by this wrapper and the Mod9 Engine.
    for google_key in google_config_settings:
        if google_key not in settings.google_allowed_keys:
            raise KeyError(f"Option key '{google_key}' not supported.")

    # Translate google_config_settings keys to corresponding Mod9 keys and values.
    for google_key, google_value in google_config_settings.items():
        mod9_key = settings.google_to_mod9_key_translations[google_key]
        # A subset of possible Google values are supported by this wrapper and the Mod9 Engine.
        if google_value in settings.mod9_keys_to_allowed_values[mod9_key]:
            # Some Mod9 values are equivalent to Google values. Others to be translated later.
            mod9_config_settings[mod9_key] = google_value
        else:
            raise KeyError(f"Option value '{google_value}' not supported for key '{google_key}'.")

    # Do first step of translating Mod9 'encoding' value from Google to Mod9 format + encoding.
    # Format will map to 'wav' or 'raw'. Set placeholder until determined by file header.
    mod9_config_settings['format'] = None
    if 'encoding' in mod9_config_settings:
        if mod9_config_settings['encoding'] == 'ENCODING_UNSPECIFIED':
            mod9_config_settings['format'] = 'wav'
            del mod9_config_settings['encoding']
        else:
            mod9_config_settings['encoding'] = \
                settings.google_encoding_to_mod9_encoding[mod9_config_settings['encoding']]

    # Set N-best settings.
    #  ``result_from_mod9()`` iterates through alternatives.
    n_best_N = mod9_config_settings.get('transcript-alternatives')
    if n_best_N is not None:
        if n_best_N == 0:
            # Google sets n_best_N: 0 -> 1.
            n_best_N = 1
        mod9_config_settings['transcript-alternatives'] = n_best_N

    # This option only applies with speech_mod9.
    if mod9_config_settings.get('phrase-alternatives'):
        mod9_config_settings['phrase-intervals'] = True

    # Set appropriate model based on user input ``languageCode``.
    if not google_config_settings.get('languageCode'):
        google_config_settings['languageCode'] = 'en'
    models = utils.find_loaded_models_with_language(google_config_settings['languageCode'])
    if len(models) == 0:
        raise ValueError(f"Language {google_config_settings['languageCode']} not supported or "
                         f"loaded. Currently loaded: {sorted(utils.get_model_languages())}")
    elif len(models) == 1:
        mod9_config_settings['asr-model'] = models[0]['name']
    else:
        model = utils.select_best_model_for_language_code(
            models,
            google_config_settings['languageCode'],
            model_type=google_config_settings.get('model'),
        )
        mod9_config_settings['asr-model'] = model['name']
    if mod9_config_settings.get('language-code'):
        del mod9_config_settings['language-code']
    if mod9_config_settings.get('model'):
        del mod9_config_settings['model']

    # These speech_mod9 options override all prior options that might have been set.
    if 'options_json' in mod9_config_settings:
        try:
            extra_options = json.loads(mod9_config_settings['options_json'])
        except Exception:
            raise ValueError('Could not parse additional options JSON.')
        mod9_config_settings.update(extra_options)
        del mod9_config_settings['options_json']

    return mod9_config_settings


def google_audio_settings_to_mod9(google_audio_settings):
    """
    Map from Google-style audio input to Mod9 TCP server-style audio
    input.

    Args:
        google_audio_settings (dict):
            Google-style audio.

    Returns:
        Union[str, Iterable[bytes]]:
            Mod9-style audio to pass to Mod9 ASR Engine TCP Server.
    """

    if not google_audio_settings:
        return None

    # Require one, and only one, of 'uri' or 'content'.
    if 'uri' in google_audio_settings and 'content' in google_audio_settings:
        raise ConflictingGoogleAudioSettingsError("Got both 'uri' and 'content' keys.")
    if 'uri' not in google_audio_settings and 'content' not in google_audio_settings:
        raise KeyError("Got neither 'uri' nor 'content' key.")

    if 'uri' in google_audio_settings:
        mod9_audio_settings = google_audio_settings['uri']
    else:
        # Decode google_audio_settings byte string if Base64 encoded; send chunks in generator.
        try:
            byte_string = base64.b64decode(google_audio_settings['content'], validate=True)
        except BinasciiError:
            byte_string = google_audio_settings['content']
        mod9_audio_settings = (
            byte_string[i:i+CHUNKSIZE] for i in range(0, len(byte_string), CHUNKSIZE)
        )

    return mod9_audio_settings


def result_from_mod9(mod9_results, logger=None):
    """
    Map from Mod9 TCP server-style output to Google-style output.

    Args:
        mod9_results (Iterable[dict]):
            Mod9-style results from the Mod9 ASR Engine TCP Server.

    Yields:
        dict:
            Google-style result.
    """

    # Set language below based on first Engine reply. Raise exception if cannot set.
    language_code = None

    # Longer audio comes chopped into segments.
    for mod9_result in mod9_results:
        if mod9_result['status'] != 'processing':
            # Non-'processing' status -> failure or transcription is complete.
            if mod9_result['status'] == 'failed':
                raise Mod9EngineFailedStatusError(f"Mod9 server issues 'failed': {mod9_result}.")
            elif mod9_result['status'] != 'completed':
                if logger:
                    logger.error("Unexpected Mod9 server response: %s.", mod9_result)
                else:
                    logging.error("Unexpected Mod9 server response: %s.", mod9_result)
            else:
                # Status 'completed' is final response (with no transcript).
                break

        if 'result_index' not in mod9_result:
            asr_model = mod9_result.get('asr_model')
            if asr_model:
                models = utils.get_loaded_models_mod9()
                for model in models:
                    if model['name'] == asr_model:
                        language_code = model['language']
                        break
                if not language_code:
                    raise utils.Mod9EngineCouldNotGetModelNameError(
                        'Could not set language code '
                        f"(asr_model field not found in {mod9_result})."
                    )
            continue

        alternatives = []
        if 'alternatives' in mod9_result:
            for alternative_number, mod9_alternative in enumerate(mod9_result['alternatives']):
                alternative = build_google_alternative(mod9_result['result_index'])
                if alternative_number == 0 and 'transcript_formatted' in mod9_result:
                    alternative['transcript'] += mod9_result['transcript_formatted']
                else:
                    alternative['transcript'] += mod9_alternative['transcript']
                alternatives.append(alternative)
        else:
            # Partial results (``.final`` == ``False``) from Mod9 do not have alternatives
            #  or ``.transcript_formatted``.
            #  Requests without transcript alternatives do not have alternatives,
            #  but may have ``.transcript_formatted``.
            alternative = build_google_alternative(mod9_result['result_index'])
            if 'transcript_formatted' in mod9_result:
                alternative['transcript'] += mod9_result['transcript_formatted']
            else:
                alternative['transcript'] += mod9_result['transcript']
            alternatives.append(alternative)

        # Build the WordInfo if Mod9 has returned word-level results.
        #  If returning N-best, only 1-best gets word alternatives.
        if 'words' in mod9_result:
            words = []
            for mod9word in mod9_result['words']:
                new_word = OrderedDict()
                if 'interval' in mod9word:
                    start_time, end_time = mod9word['interval']
                    new_word['startTime'] = "{:.3f}s".format(start_time)
                    new_word['endTime'] = "{:.3f}s".format(end_time)

                new_word['word'] = mod9word['word']

                if 'confidence' in mod9word:
                    new_word['confidence'] = mod9word['confidence']

                words.append(new_word)
            alternatives[0]['words'] = words

        google_result = OrderedDict(
            [
                ('alternatives', alternatives),
                ('isFinal', mod9_result['final']),
                # NOTE: this is only returned in v1p1beta1, and it's lowercase for some reason.
                ('languageCode', language_code),
                ('resultEndTime', "{:.3f}s".format(mod9_result['interval'][1]))
            ]
        )
        if not mod9_result['final']:
            google_result['stability'] = 0.0  # Google's default.

        if 'phrases' in mod9_result:
            google_result['phrases'] = mod9_result['phrases']
            for phrase in google_result['phrases']:
                interval = phrase.pop('interval')
                phrase['startTime'] = "{:.3f}s".format(interval[0])
                phrase['endTime'] = "{:.3f}s".format(interval[1])

        yield google_result


def build_google_alternative(transcript_number, confidence_value=1.0):
    """
    Build template for Google alternative.

    Args:
        transcript_number (int):
            Indicate the segment/endpoint the alternative is a part of.
        confidence_value (float):
            Rating in [0.0, 1.0] indicating confidence this alternative
            is the true transcript/one-best (default is 1.0).

    Returns:
        dict:
            Google-style transcript alternative (i.e. one of N-best list).
    """

    alternative = OrderedDict([('transcript', '')])
    # Google transcripts after the first start with a space.
    if transcript_number > 0:
        alternative['transcript'] += ' '

    # Add placeholder value for transcript-level confidence.
    alternative['confidence'] = confidence_value

    return alternative


def google_type_result_from_dict(
        google_result_dicts,
        google_result_type,
        module,
):
    """
    Convert dict-type result iterable to Google-type result generator.

    Args:
        google_result_dicts (Iterable[dict]):
            Google-style results.
        google_result_type (Union[module.RecognizeResponse, module.StreamingRecognizeResponse]):
            Google-like-type to return.
        module (module):
            Module to read Google-like-types from, in case of
            subclassing.

    Yields:
        Union[module.RecognizeResponse, module.StreamingRecognizeResponse]:
            Google-like-type result.
    """

    for google_result_dict in google_result_dicts:
        if google_result_type == module.SpeechRecognitionResult:
            # ``from_json()`` complains if attributes that do not exist in a protobuf are passed.
            if 'isFinal' in google_result_dict:
                google_result_dict.pop('isFinal')
            if 'resultEndTime' in google_result_dict:
                google_result_dict.pop('resultEndTime')
        # Internal camelCase keys are converted to snake_case by from_json().
        #  See docstring in GoogleConfigurationSettingsAndMappings for more info.
        yield google_result_type.from_json(json.dumps(google_result_dict))



"""
Provides defaults used throughout mod9-asr.
"""

import logging
import os

# Current wrappers version.  Note that this is not the same as the Engine version.
WRAPPER_VERSION = '1.3.0'

# CHANGELOG:
#   1.3.0 (18 Nov 2021):
#   - Add mod9-asr-elasticsearch-client to demonstrate indexing of phrase alternatives.
#   - Enable non-English languages to be specified with "languageCode" option.
#     Unlike Google STT, a region suffix may be omitted, e.g. "en" instead of "en-US".
#   - Support the "model" option, similarly to Google STT.
#   1.2.1 (11 Nov 2021):
#   - Bugfix to allow WebSocket server to handle responses up to 1 MiB (instead of 64KiB).
#   - This setting may be overriden with the --websocket-limit-bytes option.
#   1.2.0 (30 Aug 2021):
#   - Improved logging.
#   - Allow "rate" option to be in the range [8000,48000], as with Google STT.
#   - Added "speed" option to speech_mod9.
#   - Added "options_json" to speech_mod9.
#   1.1.1 (11 Aug 2021):
#   - Rebuild correctly (after `rm -rf build/ dist/ *.egg-info`)
#   1.1.0 (11 Aug 2021):
#   - Released in coordination with Engine version 1.1.0 (coincidental version match, not causal).
#   - Added "latency" request option to speech_mod9.
#   - REST API now logs to a file, with UUIDs both for itself and the proxied Engine.
#   1.0.0 (31 Jul 2021):
#   - This version is not compatible with Engine version < 1.0.0 (due to "asr-model" option).
#   - Bugfixes to WebSocket interface; also add --skip-engine-check and --allow-*-uri (for REST).
#   0.5.0 (28 May 2021): Add Websocket Interface.
#   0.4.1 (20 May 2021): Additional minor documentation fixes; Flask-RESTful version pinning.
#   0.4.0 (30 Apr 2021): Rename mod9-rest-server to mod9-asr-rest-api; minor documentation fixes.

# Range of compatible Engine versions for current wrappers.
#  Lower bound is inclusive, upper bound is exclusive.
#  ``None`` indicates no bound.
WRAPPER_ENGINE_COMPATIBILITY_RANGE = ('1.0.0', None)  # tested at 1.2.0 as of 2021 Aug 29.

ASR_ENGINE_HOST = os.getenv('ASR_ENGINE_HOST', 'localhost')
ASR_ENGINE_PORT = int(os.getenv('ASR_ENGINE_PORT', 9900))

SOCKET_CONNECTION_TIMEOUT_SECONDS = 10.0
SOCKET_INACTIVITY_TIMEOUT_SECONDS = 60.0
ENGINE_CONNECTION_RETRY_SECONDS = 1.0

MAX_CHUNK_SIZE = 8 * 1024 * 1024  # Used as chunk size for URI producers; limits generators.
GS_CHUNK_SIZE = 262144  # Google requires chunks be multiples of 262144

FLASK_ENV = os.getenv('FLASK_ENV', None)

# Audio URI prefixes to accept, used by REST only (PySDK allows all).
#  Operator can set at server launch; default is allow none.
ASR_REST_API_ALLOWED_URI_SCHEMES = os.getenv('ASR_REST_API_ALLOWED_URI_SCHEMES', set())
if ASR_REST_API_ALLOWED_URI_SCHEMES:
    ASR_REST_API_ALLOWED_URI_SCHEMES = ASR_REST_API_ALLOWED_URI_SCHEMES.lower().split(sep=',')
    ASR_REST_API_ALLOWED_URI_SCHEMES = set(
        scheme.replace('://', '') for scheme in ASR_REST_API_ALLOWED_URI_SCHEMES
    )

if 'http' in ASR_REST_API_ALLOWED_URI_SCHEMES and 'https' not in ASR_REST_API_ALLOWED_URI_SCHEMES:
    logging.warning('REST API set to allow http:// but NOT https:// audio URIs.')

# Limit on number of bytes allowed per reply line read by WebSocket server.
WEBSOCKET_LIMIT_BYTES = 1024 * 1024  # 1 MiB



#!/usr/bin/env python3

import sys

import json


def engine_phrases_to_es_cats(json_lines):
    concats = []
    for json_line in json_lines:
        engine_reply = json.loads(json_line)
        if engine_reply.get('final'):
            concats += engine_reply.get('phrases', '')

    return {'cats': concats}
    # return {'cats': json.dumps(concats)}


def main():
    print(json.dumps(engine_phrases_to_es_cats(sys.stdin)))


if __name__ == '__main__':
    main()
